# dss.pl
#
# ---------------------------------------------------------------------
# By Joost "Garion" Vunderink
# This file is part of spamcalc.
# See http://spamcalc.net/ for more information.
# ---------------------------------------------------------------------
#
# Tool to automatically deny clients access to your server if they connect
# with a dnsspam hostname.
#
# IMPORTANT NOTE:
#
# This script requires SpamCalc.pm to be in ~/.irssi/scripts/spamcalc, 
# and the spamcalc datafiles in ~/.irssi/scripts/spamcalc/data
# Without these files, the script doesn't work (properly).
# 
# ---------------------------------------------------------------------

# Commands in control channel:
#
# These are either
# !<command> <args>
# or
# nick: <command> <args>
# The first one triggers all spamcalc clients, the second one only <nick>.
# 
# !sc <host>
#   calculates score
# !scv <host>
#   shows the verbose calculation on host
# !version
#   shows version of script and module
# !stats
#   shows some statistics
# !info
#   shows information about the operational status
# !fetch       
#   gets most recent datafiles
# !reload
#   reloads the datafiles
# !nb [show|clear|clearall]
#   shows, or clears the netblock data
#
# !trace <show|kline>
#   traces all users on your server and either SHOWs or KLINEs the hosts
#   default is "show" of course. Only works on EFnet.
#

# Issues are with
# a) 1 spamcalc client on multiple servers
# b) multiple spamcalc clients on the same channel.

# Well, b) shouldn't be a problem, as long as you make sure only one of
# the clients answers to !sc and !info. The rest can be processed by all
# those clients (if necessary, with access based on hostmask).

# a) might be a bigger problem. If we allow spamcalc to run on multiple
# servers, we'll have to use settings per server - 1 setting for on/off
# or the penalty score doesn't suffice then.
# So, that will require lots of settings parsing. I don't really feel
# like that.

# Multiserver stuff will probably be restricted to being able to set
# the reporting channel on some other network.

# ---------------------------------------------------------------------
# Test if the spamcalc module is present

my $found_again_module;
my $module_file;
use Irssi;

BEGIN {
	$module_file = Irssi::get_irssi_dir() . "/scripts/spamcalc/SpamCalc.pm";
	if (!-f $module_file) {
		Irssi::print("SpamCalc.pm not found. Error loading dss.pl.");
		return 0;
	}

	$found_again_module = eval { require again; };
	unshift @ISA, Irssi::get_irssi_dir() . "/scripts";
	if (eval { require again }) {
		Irssi::print("%ydss%n %_again%_ found. If spamcalc::SpamCalc is modified, ".
			         "just reload this script and the module will be reloaded.", MSGLEVEL_CRAP);
		again::require_again('spamcalc::SpamCalc');
		Irssi::print("%ydss%n SpamCalc.pm loaded.", MSGLEVEL_CRAP);
	} else {
		Irssi::print("%ydss%n %_again%_ not found. If spamcalc::SpamCalc is modified ".
			         "you must restart irssi and load this script for the changes. ".
			         "Use perl -MCPAN -e'install again' to install the again perl ".
			         "module or get again.pm from http://spamcalc.net/ to solve this ".
			         "problem.", MSGLEVEL_CRAP);
		require spamcalc::SpamCalc;
		Irssi::print("%ydss%n SpamCalc.pm loaded.", MSGLEVEL_CRAP);
	}
}

my $irssidir = Irssi::get_irssi_dir();
my $datafilesdir = $irssidir . "/scripts/spamcalc/data";

# ---------------------------------------------------------------------
# Main script stuff

use strict;
use vars qw($VERSION %IRSSI);

($VERSION) = '$Revision: 1.6 $' =~ / (\d+\.\d+)/;
%IRSSI = (
		authors	=> 'Garion',
		contact	=> 'garion@efnet.nl',
		name	=> 'dss',
		description	=> 'Tools for (automatic) calculation and denial of dnsspam hostnames on an irc-server.',
		license	=> 'Public Domain',
		url		=> 'http://spamcalc.net/',
		changed	=> '14 Aug 2002 00:34:38',
);

# ---------------------------------------------------------------------
# Globals
 
my $calc;		# the calculator
my %stats;
my %prefs;

my %netblock;

my $highscore = 0;		# the score of the highest spamhost detected
my $highhost = 0;		# the hostname that scored most
my $numpossible = 0;	# the number of possible spam connections
my $numcertain = 0;		# the number of certain spam connections
my $numklines = 0;		# the number of placed k-lines
my $numconnections = 0;	# the number of connections tested
my $starttime;			# the time and date the script was loaded
my $totalspam = 0;		# the total amount of points scored (-> average)
my $numzeroes = 0;		# the number of connections that scored 0
my $numwhitelist = 0;	# the number of connections that were whitelisted

# Some ugly global vars to deal with the !trace command.
my $tracekline = 0;
my $traceklinechannel = "";
my $traceklinenum = 0;
my $traceklinenumlusers = 0;
my $traceklineshowonly = 1;

# Deals with the periodic downloading of spamcalc datafiles.
my $delta_handle_fetch;
my $fetch_delta;


# ----------------------------------------------------------------------
# A Server Event has occurred. Check if it is a server NOTICE;
# if so, process it.

sub event_serverevent {
    my ($server, $msg, $nick, $hostmask) = @_;
    my ($nickname, $username, $hostname);

    # If it is not a NOTICE, we don't want to have anything to do with it.
    if ($msg !~ /^NOTICE/) {
        return;
    }

    # If the hostmask is set, it is not a server NOTICE, so we'll ignore it
    # as well.
    if (length($hostmask) > 0) {
        return;
    }

    my $ownnick = $server->{'nick'};

    process_server_notice($server, $msg);
}

# ----------------------------------------------------------------------
# Check whether the incoming server notice is a client connect notice.

sub process_server_notice {
	my ($server, $msg) = @_;

	# client_connect ($server, $nick, $user, $host, $ip, $real)

	# HYBRID 7 - need a setting to determine which server type we have!
	# Client connect: nick, user, host, ip, class, realname
	if ($msg =~ /Client connecting: (.*) \((.*)@(.*)\) \[(.*)\] {(.*)} \[(.*)\]/) {
		client_connect($server, $1, $2, $3, $4, $6);
		return;
	}

	# HYBRID 6
	# Client connect: nick, user, host, ip, class [no realname! gasp!]
	if ($msg =~ /Client connecting: (.*) \((.*)@(.*)\) \[(.*)\] {(.*)}/) {
		client_connect($server, $1, $2, $3, $4, "");
		return;
	}

	# ratbox
	# Client connecting: Hen-P` (ooasr@bzq-193-233.red.bezeqint.net) [212.179.193.233] {us} [DevilmaN]
	# same as hyb7.
	#
	# csircd 1.14
	# [7:55am] Client connecting: irssBen (royalsexy@oldschool.ircd.rocks) [127.0.0.1] {ozdweller}
	# same as hyb6.

	# IRCnet
	# Client connecting: Pilah is none from ca-marseille-42-159.w80-8.abo.wanadoo.fr [80.8.169.159] named Philippe
	if ($msg =~ /Client connecting: (.*) is (.*) from (.*) \[(.*)\] named (.*)/) {
		client_connect($server, $1, $2, $3, $4, $5);
		return;
	}
}

# ---------------------------------------------------------------------
# Process the client connection.

sub client_connect {
    my ($server, $nick, $user, $host, $ip, $real) = @_;
	$numconnections++;
    my $now = time();
	my $reportchan = Irssi::settings_get_str('dss_report_channel');
	
	# Don't do anything for unresolved ipv6 ips
	if ($host =~ /:/) {
		return;
	}

	# Don't do anything for unresolved ipv4 ips
	if ($host =~ /[0-9]$/) {
		return;
	}

	my $score = $calc->get_host_score($host);
	$totalspam += $score;
	if ($score == 0) {
		$numzeroes++;
	}

	cmd_log_host($score . " - " . $host);

	dprint(2, "Connect: $score - $host ($nick).");

	if ($score >= Irssi::settings_get_int('dss_threshold_certain')) {
		process_client_spam($server, $nick, $user, $host, $ip, $score);
		return;
	}
	
	if ($score >= Irssi::settings_get_int('dss_threshold_possible')) {
		process_client_possible($server, $nick, $user, $host, $ip, $score);
	    return;
	}
}

# ---------------------------------------------------------------------
# Takes action on a dnsspam client: report it to the channel, and 
# possibly deny it access to the server.

sub process_client_spam {
	my ($server, $nick, $user, $host, $ip, $score) = @_;

	add_to_netblock($server, $ip);
	$numcertain++;

	# Update the highscore data if necessary.
	if ($score > $highscore) {
  		$highscore = $score;
		$highhost = $host;
	}

	if (Irssi::settings_get_bool('dss_enable')) {
		dprint(1, "$score client on host $host denied.");
		deny_client($server, $host);
		report_action($server, "kline", $nick, $host, $ip, $score);
		cmd_log_kline("K-line added: $score - $host.");
	} else {
		dprint(1, "$score client on host $host logged.");
		report_action($server, "spam", $nick, $host, $ip, $score);
		cmd_log_kline("Logged: $score - $host.");
	}
}	

# ---------------------------------------------------------------------
# Adds this ip to the netblock hash in the right netblock. For ipv4,
# its /24 is used, and for ipv6 its /48 is used.
# Also reports a warning to the reporting channel if this causes the
# number of dnsspam connections in that netblock to be above the min
# value dss_netblock_min.
# TODO: change this into /64. Need some decent code for that.

sub add_to_netblock {
	my ($server, $ip) = @_;
	my $protocol = 4;
	if ($ip =~ /:/) { $protocol = 6; }

	my $nb = $ip;
	if ($protocol == 4) {
		$nb =~ s/\.[0-9]+$//;
	} else {
		$nb =~ s/^([0-9a-f]+:[0-9a-f]+:[0-9a-f]+::?).*/\1/;
	}
	$netblock{$nb}++;

	dprint(0, "Added $ip to netblock nb which now has " . $netblock{nb} .
			" dnsspam connections.");
	
	if ($netblock{$nb} > Irssi::settings_get_int('dss_netblock_min') &&
			Irssi::settings_get_bool('dss_netblock_report_warning')) {
		send_chan_msg($server, Irssi::settings_get_str('dss_report_channel'),
			"Warning: $nb" . ($protocol == 4 ? "." : ":") . "* has had " . 
				$netblock{$nb} . " dnsspam connections.");
	}
}

# ---------------------------------------------------------------------
# Processes a possible spam client, adding them to the possible spam
# count, and reporting this to the reporting channel.

sub process_client_possible {
	my ($server, $nick, $user, $host, $ip, $score) = @_;
	dprint(1, "$score client on host $host logged.");
	$numpossible++;
	report_action($server, "poss", $nick, $host, $ip, $score);
}
	
# ---------------------------------------------------------------------
# Places the correct line so that the client with host $host is
# denied access to the server.

sub deny_client {
	my ($server, $host) = @_;
	my $servertype = Irssi::settings_get_str('dss_server_type');
	my $time = Irssi::settings_get_int('dss_time');
	my $reason = Irssi::settings_get_str('dss_reason');

	my $cmd;
	if ($servertype eq "undernet") {
		my $glinetime = $time * 60;
		$cmd = "GLINE +*\@$host $glinetime :$reason";
		dprint(0, "Placed local G-line ($time mins) on $host");
	} elsif ($servertype eq "ircnet") {
		$cmd = "";
		# nothing!
	} else {
		dprint(0, "Placed temp K-line ($time mins) on $host");
		$cmd = "KLINE $time *\@$host :$reason";
	}

	if (length($cmd) > 0 ) {
		$server->send_raw($cmd);
	}
}

# ---------------------------------------------------------------------
# Reports an action to the reporting channel.

sub report_action {
	my ($server, $action, $nick, $host, $ip, $score) = @_;

	my $channel = Irssi::settings_get_str('dss_report_channel');
	my $msg;

	if ($action eq "kline") {
		$msg = sprintf("\002Kline\002 %4d [%15s %9s] \002%s\002", $score, align_ip($ip), $nick, $host);
	} elsif ($action eq "spam") {
		$msg = sprintf("Spam  %4d [%15s %9s] \002%s\002", $score, align_ip($ip), $nick, $host);
	} elsif ($action eq "poss") {
		$msg = sprintf("Poss  %4d [%15s %9s] %s", $score, align_ip($ip), $nick, $host);
	} elsif ($action eq "trace") {
		$msg = sprintf("Trace %4d [%15s %9s] \002%s\002", $score, align_ip($ip), $nick, $host);
	} elsif ($action eq "tracekline") {
		$msg = sprintf("\002Tdeny\002 %4d [%15s %9s] \002%s\002", $score, align_ip($ip), $nick, $host);
	} else {
		$msg = sprintf("Unkown action $action.");
	}

	send_chan_msg($server, $channel, $msg);
}

# ---------------------------------------------------------------------
# Sends $msg to $chan if both are at least 1 character long.

sub send_chan_msg {
	my ($server, $chan, $msg) = @_;
	if (length($chan) > 0 && length($msg) > 0) {
		$server->command("msg $chan $msg");
	}
}

# ---------------------------------------------------------------------
# Appends spaces to this IP to make it 15 characters long.

sub align_ip {
	my ($ip) = @_;
	return $ip . " " x (15 - length($ip));
}

# ---------------------------------------------------------------------
# Checks for public !command queries
# TODO: cleanup this function

sub event_public {
	my ($server, $data, $nick, $mask, $channel) = @_;
	my $legal = is_on_acl($mask);
	my $ownnick = $server->{nick};
	my $command;

	if ($data =~ /^$ownnick[:,]?\s+(.+)/i) {
		$command = $1;
	} elsif ($data =~ /^!(.+)/) {
		$command = $1;
	} else {
		return;
	}
	# Strip spaces from the command
	$command =~ s/^\s+//;
	$command =~ s/\s+$//;

	# TODO: Test here whether the request is on a request enabled channel

	if ($command =~ /^statu?s/i) {
		process_event_stats($server, $channel);
		
	} elsif ($command =~ /^version/i) {
		process_event_version($server, $channel);

	} elsif ($command =~ /^info/i ) {
		process_event_info($server, $channel);

	} elsif ($command =~ /^(spamcalc|spam|sc|calc) ([0-9a-zA-Z_\.-]+)$/) {
		process_event_calc($server, $channel, $2);
		
	} elsif ($command =~ /^scv ([0-9a-zA-Z_\.-]+)$/) {
		process_event_scv($server, $channel, $1);
		
	} elsif ($command =~ /^(word) ([0-9a-zA-Z-]+)$/ ) {
		if (Irssi::settings_get_bool('dss_enable_query_public') == 0) { return; }
		my $pen = $calc->get_word_score($2);
		$server->command("msg $channel $pen - $2.");
		
	} elsif ($command eq "nb") {
			process_event_nb($server, $channel, "", $nick);

	} elsif ($command =~ /nb\s+(.+)/) {
			process_event_nb($server, $channel, $1, $nick);

	} elsif ($command =~ /^acl/ ) {
			process_event_acl($server, $channel, $nick, $mask);

	}

# Below here are the private functtions
# You need to be cool to use these.
	if (!$legal) { return; }


	if ($command =~ /^trace/) {
		if ($server->{'usermode'} !~ /o/) {
			send_chan_msg($server, $channel, "Dude, that doesn't work if I'm not opered.");
			return;
		}
		$tracekline = 1;
		$traceklinechannel = $channel;
		$traceklinenum = 0;
		$traceklineshowonly = 1;
		if ($command =~ /kline/) {
			$traceklineshowonly = 0;
		}
		my $msg;
		if ($traceklineshowonly) {
			dprint(0, "Tracekline requested by $nick");
			$msg = "Checking all users on " .$server->{real_address}. " for dnsspam...";
		} else {
			dprint(0, "Trace requested by $nick");
			$msg = "Klining all dnsspam hostnames on " . $server->{real_address} . "...";
		}
		send_chan_msg($server, $channel, $msg);
		$server->redirect_event('scantrace', 1, '', -1, undef,
						{
						"event 204" => "redir scantrace",
						"event 205" => "redir scantrace",
						"event 262" => "redir scantracestop",
						"" => "event empty"
						});

		$server->command("TRACE");
		return;
	}

# Enable/disable
	if ($command =~ /^disable/ && $legal == 1) {
			dprint(0, "Automatic k-lines disabled by $nick.");
			$server->command("msg $channel Automatic k-lines disabled.");
			Irssi::settings_set_bool('dss_enable', 0);
	}
	if ($command =~ /^enable/ && $legal == 1) {
			dprint(0, "Automatic k-lines enabled by $nick.");
			$server->command("msg $channel Automatic k-lines enabled.");
			Irssi::settings_set_bool('dss_enable', 1);
	}

# Rehash
	if ($command =~ /^(reload|rehash)/ && $legal == 1) {
			dprint(0, "Config rehashed by $nick.");
			$calc->load_datafiles($datafilesdir);
			$server->command("msg $channel Data files reloaded from $datafilesdir");
	}

# Fetch
	if ($command =~ /^fetch/ && $legal == 1) {
			dprint(0, "Fetch of datafiles requested by $nick.");
			fetch_datafiles($server, $channel);
#$server->command("msg $channel Data files reloaded from $datafilesdir");
	}

# update
	if ($command =~ /^update/i && $legal == 1) {
			dprint(0, "Update requested by $nick.");
			if ($found_again_module) {
					again::require_again('spamcalc::SpamCalc');
					$server->command("msg $channel Reloading SpamCalc.pm... new version " . $calc->VERSION);
			} else {
					$server->command("msg $channel again.pm not loaded - can't update.");
			}
	}
}

# ---------------------------------------------------------------------
# Processes the public !nb command, showing and deleting the netblocks
# that have multiple dnsspam hosts.

sub process_event_nb {
	my ($server, $channel, $cmd, $nick) = @_;
	my %amounts;
	my $msg;
	my $nb;

	if ($cmd eq "clear" || $cmd eq "clean" || $cmd eq "del") {
		my $min = Irssi::settings_get_int('dss_netblock_min');
		foreach $nb (keys %netblock) {
			if ($netblock{$nb} >= $min) {
				delete $netblock{$nb};
			}
		}
		dprint(0, "Netblocks with more than " . ($min - 1) . " dnsspam hosts deleted.");
		send_chan_msg($server, $channel, "Netblocks with more than " . ($min - 1) . " dnsspam hosts deleted.");
		return;
	}
	
	if ($cmd eq "clearall" || $cmd eq "cleanall" || $cmd eq "delall") {
		dprint(0, "All netblocks cleared by $nick.");
		undef(%netblock);
		send_chan_msg($server, $channel, "All netblocks deleted.");
		return;
	}

	if ($cmd eq "show") {
		dprint(0, "Show all netblocks requested by $nick.");
		my $min = Irssi::settings_get_int('dss_netblock_min');
		my $total = 0;
		foreach $nb (sort {$a <=> $b} keys %netblock) {
			if ($netblock{$nb} >= $min) {
				$total++;
				$msg = $netblock{$nb} . " - " . $nb . ".*";
				send_chan_msg($server, $channel, $msg);
			}
		}
		send_chan_msg($server, $channel, "$total netblocks with $min or more dnsspam hostnames.");
		return;
	}

	
	$msg = "Checking amount of spam per netblock.";
	send_chan_msg($server, $channel, $msg);
	
	foreach my $nb (keys %netblock) {
		$amounts{ $netblock{$nb} } ++;
	}
	
	my $total = 0;
	foreach my $key (sort {$a <=> $b} keys %amounts) {
		$msg = "Netblocks with $key spamhosts: $amounts{$key}";
		$total += $amounts{$key};
		send_chan_msg($server, $channel, $msg);
	}

	dprint(0, "Show netblocks requested by $nick.");
	$msg = "Total amount of netblocks: $total.";
	send_chan_msg($server, $channel, $msg);
}

# ---------------------------------------------------------------------
# Reports statisticts to the reporting channel.

sub process_event_stats {
	my ($server, $channel) = @_;
	my $timediff = time() - $starttime;
	my $connpersec = 0;
	if ($timediff > 0) {
		$connpersec = int(100 * $numconnections / $timediff) / 100.0;
	}
	my ($wlprc, $zrprc) = (0, 0);
	if ($numconnections > 0) {
		$wlprc = int(1000 * $numwhitelist / $numconnections + 0.5) / 10.0;
		$zrprc = int(1000 * $numzeroes / $numconnections + 0.5) / 10.0;
	}
	my $wlzr = $wlprc + $zrprc;
	
	my $msg = "Past $timediff seconds: $numconnections connections ($connpersec per second).";
	send_chan_msg($server, $channel, $msg);

	$msg = "K-lines: $numklines. Spam: $numcertain. Possible: $numpossible. ";
	send_chan_msg($server, $channel, $msg);

	$msg = "Whitelisted: $numwhitelist ($wlprc\%). Zero: $numzeroes ($zrprc\%). Total: $wlzr\%.";
	send_chan_msg($server, $channel, $msg);

	if ($highscore > 0) {
		$msg = "Highest score: $highscore by $highhost.";
		send_chan_msg($server, $channel, $msg);
	}
	return;
}

# ---------------------------------------------------------------------
# Reports the version to the reporting channel.

sub process_event_version {
	my ($server, $channel) = @_;
	my $msg = "dss version $VERSION, SpamCalc module version " . $calc->VERSION;
	send_chan_msg($server, $channel, $msg);
}

# ---------------------------------------------------------------------
# Reports the score of a host to the reporting channel.

sub process_event_calc {
	my ($server, $channel, $host) = @_;

	if (Irssi::settings_get_bool('dss_enable_query_public') == 0) { return; }

	if ($host !~ /[0-9a-zA-Z.-]+/) {
		send_chan_msg($server, $channel, "Please use !sc <valid hostname>.");
		return;
	}

	my $info = $calc->get_host_info($host);
	my $score = $info->{score};
	my $msg;
	if ($score >= 100) {
			$msg = "\00304";
	} elsif ($score >= 50) {
			$msg = "\00308";
	} else {
			$msg = "\00309";
	}
	$msg .= $score . "\003 - $host";
	if ($info->{iswhitelisted}) {
			$server->command("msg $channel [whitelist] $msg");
		} else {
			$server->command("msg $channel $msg");
		}
	return;
}

# ---------------------------------------------------------------------

sub process_event_info {
	my ($server, $channel) = @_;
	$server->command("msg $channel dss.pl version $VERSION; SpamCalc.pm version " . $calc->VERSION);
	$server->command("msg $channel " .
		"words: " . $calc->get_num_words() .
		"; regexps: " . $calc->get_num_regexps() .
		"; blacklist: " . $calc->get_num_blacklist() .
		"; whitelist: " . $calc->get_num_whitelist());
	
	my $status;
	if (Irssi::settings_get_bool('dss_enable')) {
		$status .= "Automatic k-lines enabled. ";
	} else {
		$status .= "Automatic k-lines disabled. ";
	}

	my $thrposs = Irssi::settings_get_int('dss_threshold_possible');
	$status .= "Possible spam threshold: $thrposs. ";
	my $thrcert = Irssi::settings_get_int('dss_threshold_certain');
	$status .= "Certain spam threshold: $thrcert.";

	send_chan_msg($server, $channel, $status);

	# TODO: add message about (not) having right usermodes
	return;
}

# ---------------------------------------------------------------------

sub process_event_scv {
	my ($server, $channel, $host) = @_;
	my $info = $calc->get_host_info($host);
	my $msg;
	
	# 04 = bright red, 08 = yellow, 09 = bright green
	# 03 = dark green, 05 = dark red
	if ($info->{"score"} >= 100) {
		$msg = "\00304";
	} elsif ($info->{"score"} >= 50) {
		$msg = "\00308";
	} else {
		$msg = "\00309";
	}
	$msg .= $info->{"score"} . "\003 - " . $info->{"hostpart"} . " . " .
				  $info->{"domainpart"};
	send_chan_msg($server, $channel, $msg);
	
	if ($info->{"iswhitelisted"}) {
		send_chan_msg($server, $channel, "Whitelisted (" . $info->{"whitelistdomain"} . ").");
		return;
	}

	if ($info->{"isblacklisted"}) {
		send_chan_msg($server, $channel, "Blacklisted: \002" . $info->{"blacklistscore"} .
		     "\002 (" . $info->{"blacklistdomain"} . ").");
	}

	if ($info->{"wordscore"} > 0) {
		$msg = "";
		foreach my $field ( @{ $info->{"hostfields"} } ) {
			if ($info->{"wordscore"}->{"$field"} > 0) {
	    $msg .= $field . " \002" . $info->{"wordscore"}->{"$field"}. "\002; ";
	  }
		}
	if (length($msg) > 0) {
	  $msg = "Word fields: " . $msg;
	  $msg =~ s/; $/./;
			send_chan_msg($server, $channel, $msg);
	}
	}

	$msg = "";
	if ($info->{"twordscore"} > 0) {
		append_if_larger(\$msg, "Total word score: \002%value\002; ", $info->{"twordscore"}, 0);
		append_if_larger(\$msg, "Word multiplier: \002%value\002; ", $info->{"wordmultiplier"}, 1);
	}
	if ($info->{"tregexpscore"} > 0) {
		append_if_larger(\$msg, "Total regexp score: \002%value\002; ", $info->{"tregexpscore"}, 0);
		append_if_larger(\$msg, "Regexp multiplier: \002%value\002; ", $info->{"regexpmultiplier"}, 1);
	}
	send_chan_msg($server, $channel, $msg);

	$msg = "";
	append_if_larger(\$msg, "Numfield score: \002%value\002; ", $info->{"numfieldscore"}, 0);
	append_if_larger(\$msg, "Repeat score: \002%value\002; ", $info->{"repeatscore"}, 0);
	append_if_larger(\$msg, "Dashes score: \002%value\002; ", $info->{"dashesscore"}, 0);
	append_if_larger(\$msg, "Field length score: \002%value\002; ", $info->{"fieldlengthscore"}, 0);
	append_if_larger(\$msg, "One letter score: \002%value\002; ", $info->{"oneletterscore"}, 0);
	append_if_larger(\$msg, "Looks like word score: \002%value\002; ", $info->{"lookslikewordscore"}, 0);
	send_chan_msg($server, $channel, $msg);

	$msg = "Calculation: " . $info->{"score"} . " = " . $info->{"calculation"};
	send_chan_msg($server, $channel, $msg);

	return;
}

# ---------------------------------------------------------------------

sub process_event_acl {
	my ($server, $channel, $nick, $mask) = @_;

	if (Irssi::settings_get_bool('dss_enable_remote') == 0) {
		$server->command("msg $channel Remote control disabled.");
	} else {
	    my @remote_hostmasks = split(/ +/, Irssi::settings_get_str('dss_acl_hostmasks'));
		my $rhstring;

		# Test if the user
	    foreach my $remote_hostmask (@remote_hostmasks) {
       		if (Irssi::mask_match_address("*!".$remote_hostmask, "a", $mask)) {
       			$rhstring .= "\002$remote_hostmask\002 ";
       		} else {
       			$rhstring .= "$remote_hostmask ";
			}
   		}

		$server->command("msg $channel ACL: $rhstring");
			
	}
}

# ---------------------------------------------------------------------
# This function appends $append to $stringref if $value is larger than
# $minimum. If $string contains "%value", that is replaced by $value.

sub append_if_larger {
	my ($stringref, $append, $value, $minimum) = @_;

	if ($value <= $minimum) {
		return;
	}
	$append =~ s/%value/$value/g;
	$$stringref .= $append;
}

# ---------------------------------------------------------------------
# This is one line of /TRACE output. This function is only called if
# the event redirection has been activated, i.e. if someone has issued
# a !trace command on the reporting channel.
#
# It only works for EFnet like this. Need to investigate how to do
# this (with /WHO) on Undernet.

sub event_trace_line {
	my ($server, $msg) = @_;
	# $msg is ":<server> 205 <ownnick> User <class> <nick>[<user>@<host>] (<ip>) <num> <num>"

	$traceklinenumlusers++;

	my $ownnick = $server->{nick};

	if ($msg =~ /$ownnick User .* (.+)\[(.+)@(.+)\] \((.*)\)/) {
		my ($nick, $user, $host, $ip) = ($1, $2, $3, $4);
		my $score = $calc->get_host_score($3);
		if ($score >= Irssi::settings_get_int('dss_threshold_certain')) {
			$traceklinenum++;
			if ($traceklineshowonly) {
				report_action($server, "trace", $nick, $host, $ip, $score);
				dprint(0, "Trace found $host score $score ($nick).");
			} else {
				report_action($server, "tracekline", $nick, $host, $ip, $score);
				dprint(0, "Trace found $host score $score ($nick). K-lined.");
				my $reason = Irssi::settings_get_str('dss_reason');
				my $time = Irssi::settings_get_int('dss_time');
				if ($time == 0) { $time = "" };
				$server->send_raw("KLINE $time *\@$host :$reason");
			}
		}
	}

	Irssi::signal_stop();
}

# ----------------------------------------------------------------------
# End of /trace.

sub event_trace_end {
	my ($server, $msg) = @_;
	Irssi::print("End of /trace.");
	if ($tracekline == 1) {
		dprint(0, "Done. Found $traceklinenum hosts in $traceklinenumlusers users.");
		send_chan_msg($server, $traceklinechannel, "Done. Found $traceklinenum hosts in $traceklinenumlusers users.");
		$tracekline = 0;
		$traceklinechannel = "";
		$traceklinenum = 0;
		$traceklinenumlusers = 0;
		$traceklineshowonly = 1;
	}
	Irssi::signal_stop();
}

# ---------------------------------------------------------------------
# Calculates the spam value of a hostname and prints it.

sub cmd_dss_calc {
	my ($host) = @_;

	# If ipv6-ip, return 0
	if ($host =~ /:/) { return 0; }

	my $score = $calc->get_host_score($host);

	my $colour;
	if ($score >= Irssi::settings_get_int('dss_threshold_certain')) { 
		$colour = "%R";
	} elsif ($score >= Irssi::settings_get_int('dss_threshold_possible')) {
		$colour = "%Y";
	} else {
		$colour = "%G";
	}

	dprint(0, $colour . $score . "%n - " . $host);
}

# ---------------------------------------------------------------------
# Adds a string to the hosts logfile, if logging has been enabled.

sub cmd_log_host {
	if (Irssi::settings_get_bool('dss_enable_log_hosts') == 0) { return; }
	my ($data) = @_;
	my $time = sprintf("%02d:%02d", (localtime(time))[2], (localtime(time))[1]);
	my $msg = $time . " " . $data . "\n";
	my $file = ">>" . Irssi::settings_get_str('dss_logfile_hosts');
	open(LOGFILE, $file);
	print LOGFILE ($msg);
	close(LOGFILE);
}

# ---------------------------------------------------------------------
# Adds a string to the action logfile, if logging has been enabled.

sub cmd_log_kline {
	if (Irssi::settings_get_bool('dss_enable_log_klines') == 0) { return; }
	my ($data) = @_;
	my ($data) = @_;
	my $time = sprintf("%02d:%02d", (localtime(time))[2], (localtime(time))[1]);
	my $msg = $time . " " . $data . "\n";
	my $file = ">>" . Irssi::settings_get_str('dss_logfile_klines');
	open(LOGFILE, $file);
	print LOGFILE ($msg);
	close(LOGFILE);
}
	

# ---------------------------------------------------------------------
# Klines this hostmask with a dnsspam reason.

sub cmd_dss_kline {
	my ($host, $server, $channel) = @_;

	my $score = $calc->get_host_score($host);
	if ($score < Irssi::settings_get_int('dss_threshold_certain')) {
		dprint(0, "$score - $host: not denying this host.");
		return;
	}

	dprint(0, "Denying $host on our server.");
	deny_client($server, $host);
}

# ---------------------------------------------------------------------
# Reloads spamcalc data.

sub cmd_dss_reload {
	dprint(0, "Reloading spamcalc data...");
	$calc->load_datafiles($datafilesdir);
	dprint(0, "Done.");
}

# ---------------------------------------------------------------------
# Shows the status of the script.

sub cmd_dss_status {
	if (Irssi::settings_get_bool('dss_enable')) {
		dprint(0, "Status: Automatic k-lines enabled.");
	} else {
		dprint(0, "Status: Automatic k-lines disabled.");
	}

	my $thr_poss = Irssi::settings_get_int('dss_threshold_possible');
	dprint(0, ">= $thr_poss - possible spam."); 
	my $thr_cert = Irssi::settings_get_int('dss_threshold_certain');
	dprint(0, ">= $thr_cert - certain spam."); 
 
	my $reportchan = Irssi::settings_get_str('dss_report_channel');
	if (length($reportchan) > 0) {
		dprint(0, "K-lines will be reported to $reportchan.");
	}

	if (Irssi::settings_get_bool('dss_enable_log_hosts')) {
		my $logfile = Irssi::settings_get_str('dss_logfile_hosts');
		dprint(0, "All hosts will be logged to file \"$logfile\".");
	}

	if (Irssi::settings_get_bool('dss_enable_log_klines')) {
		my $logfile = Irssi::settings_get_str('dss_logfile_klines');
		dprint(0, "K-lines will be logged to file \"$logfile\".");
	}

	if (Irssi::settings_get_bool('dss_enable_query_public')) {
		dprint(0, "Public !sc queries enabled.");
	} else {
		dprint(0, "Public !sc queries disabled.");
	}

	if (Irssi::settings_get_bool('dss_enable_remote')) {
		my $hosts = Irssi::settings_get_str('dss_acl_hostmasks');
		dprint(0, "Remote control enabled by $hosts.");
	} else {
		dprint(0, "Remote control disabled.");
	}
}

# ---------------------------------------------------------------------
# Main dnsspam command.

sub cmd_dss {
	my ($data, $server, $item) = @_;
	if ($data =~ m/^[(calc)|(help)|(kline)|(reload)|(status)]/i ) {
		Irssi::command_runsub ('dss', $data, $server, $item);
	}
	else {
		Irssi::print("Use /dss help for help.")
	}
}

# ---------------------------------------------------------------------
# Returns 1 if hostmask $userhost is on the ACL, and 0 otherwise.
# TODO: change this so it accepts wildcards.

sub is_on_acl {
	my ($userhost) = @_;
	my @remote_hostmasks = split(/ +/, Irssi::settings_get_str('dss_acl_hostmasks'));

	# Test if the user
	foreach my $remote_hostmask (@remote_hostmasks) {
		if (Irssi::mask_match_address("*!".$remote_hostmask, "a", $userhost)) {
			return 1;
		}
	}

	return 0;
}

# ---------------------------------------------------------------------
# Fetches the datafiles from a remote location.

sub fetch_datafiles {
	my ($server, $channel) = @_;
	dprint(0, "Fetching datafiles.");
	my $fetchlocation = Irssi::settings_get_str('dss_fetch_location');
	my $fetchcmd = ("$irssidir/scripts/spamcalc/scdatafetch $datafilesdir $fetchlocation");
	
	system($fetchcmd . " >/dev/null 2&>/dev/null");

	if (defined($server)) {
		$server->command("msg $channel Fetched datafiles. Switching over.");
	}
	$calc->load_datafiles($datafilesdir);
}

# ---------------------------------------------------------------------
# If enabled, fetch the datafiles automatically.

sub delta_fetch {
	if (Irssi::settings_get_bool('dss_fetch_enable')) { 
		dprint(0, "Automatically fetching datafiles.");
		fetch_datafiles();
	}

	my $time = Irssi::settings_get_int("dss_fetch_delta") * 60 *1000;
	if ($time < 60000) { $time = 600000; }

	# Update the timer, if the delta time has changed.
	if ($time != $fetch_delta) {
		dprint(0, "Changing fetch timer from $fetch_delta to $time minutes.");
		if ($delta_handle_fetch) { Irssi::timeout_remove($delta_handle_fetch); }
		$fetch_delta = $time;
		Irssi::timeout_add($fetch_delta, 'delta_fetch', undef);
	}
}


# ---------------------------------------------------------------------
# Debug prints a message if dss_debug == true.

sub dprint {
	my ($level, $msg) = @_;
	if ($level == 0) {
		Irssi::print("%ydss%n $msg", MSGLEVEL_CRAP);
		return;
	} elsif (Irssi::settings_get_int('dss_debug_level') >= $level) {
		Irssi::print("%ydss" . $level . "%n $msg", MSGLEVEL_CRAP);
	}
}

# ---------------------------------------------------------------------
# Prints help

sub cmd_dss_help {
	Irssi::print(
"%GDNSSPAMKLINE%n - Tools to (automatically) deal with dnsspam.

%YCOMMANDS%n

%_/DSS CALC <hostname>%_
Calculates the spam value of a hostname, and prints the result.

%_/DSS KLINE <user\@hostmask>%_
Sets a temporary K-line on the given hostmask with the K-line time and reason in the settings.

%_/DSS RELOAD%_
Reloads the data of the spamcalc part. Use this after you have modified datafiles in the spamcalc dir.

%_/DSS STATUS%_
Shows some status info about this script.


%YSETTINGS%n

The script has a lot of settings to influence what's happening with it.
Here is the list:

%_dss_server_type%_
The type of server the bot is on. This can be either 'efnet', 'undernet' or 'ircnet'. For 'efnet', the bot will place temporary K-lines; for 'undernet', the bot will place local G-lines; and for 'ircnet' the bot will only report.

%_dss_enable%_
If enabled, the script will set automatic k-lines on people that connect with a hostname whose spam value exceeds dss_threshold_kline.

%_dss_enable_remote%_
Set this to 1 if you want remote control to be enabled.

%_dss_enable_log_klines%_
If enabled, all k-lines will be logged to the file in dss_logfile_klines.

%_dss_enable_log_hosts%_
If enabled, all calculations will be logged to the file in dss_logfile_hosts.

%_dss_enable_query_public%_
If enabled, the script will respond to channel messages starting with !.

%_dss_time%_
The duration of the k-line, in minutes. Will be ignored if the ircd does not support temporary k-lines.

%_dss_reason%_
The reason for the k-line.

%_dss_debug%_
Enable to show debug messages.

%_dss_logfile_hosts%_
The file where all calculations will be logged to.

%_dss_logfile_klines%_
The file where all automatic actions will be logged to.

%_dss_use_external_command (NOT YET FUNCTIONAL)%_
If enabled, instead of using /QUOTE KLINE to set k-lines, the script will call the command in dss_kline_external_command.

%_dss_external_command (NOT YET FUNCTIONAL)%_
This is the command that will be called instead of /QUOTE KLINE. If %spamvalue is in the command, it will be replaced by the spam value of the hostname; %host will be replaced by the hostname itself.

%_dss_threshold_possible%_
The value above which a hostname is possibly dnsspam.

%_dss_threshold_certain%_
The value above which a hostname is certainly dnsspam.

%_dss_acl_hostmasks%_
The space separated list of user\@hostname of all clients that are allowed to remote control this script. Use with care!

%_dss_report_channel%_
The channel where the actions of the script are reported to.

%_dss_fetch_enable%_
Set this to ON to enable automatic fetching of the spamcalc datafiles from a remote location.

%_dss_fetch_delta%_
The time (in minutes) between two consecutive data fetches.

%_dss_fetch_location%_
The url where to find the data files.

%_dss_netblock_min%_
The minimum amount of dnsspam that must be found in a single netblock (/24 for ipv4, /48 for ipv6) before a warning is reported.

%_dss_netblock_report_warning%_
Whether the above warning will be reported to the reporting channel.

%YREMOTE CONTROL%n

The script can be remotely controlled by the users that have hostmasks matching one of those in dss_acl_hostmasks. This makes it easy for you to just detach this client and remote control it, or to allow others to affect it if you are away.

Remote control is done on the reporting channel. Each command is of the form

%_!<command> <args>%_
or
%_<nick>: <command> <args>%_

The first version will issue <command> to all spamcalc clients, the second only to <nick>. The commands are:

%_version%_ shows the version
%_stats%_   shows status message about the connections spamcalc has processed
%_info%_    shows information about the spamcalc data being used
%_disable%_ disables K-lines
%_enable%_  enables K-lines
%_stats%_   shows statistics
%_sc <host>%_  shows spamcalc score of <host>
%_scv <host>%_ shows spamcalc score and calculation of <host>
", MSGLEVEL_CLIENTCRAP);
}

# ---------------------------------------------------------------------
# Add command bindings

Irssi::command_bind('dss',         'cmd_dss');
Irssi::command_bind('dss calc',    'cmd_dss_calc');
Irssi::command_bind('dss help',    'cmd_dss_help');
Irssi::command_bind('dss kline',   'cmd_dss_kline');
Irssi::command_bind('dss reload',  'cmd_dss_reload');
Irssi::command_bind('dss status',  'cmd_dss_status');


# ---------------------------------------------------------------------
# Add signal hooks

Irssi::Irc::Server::redirect_register('scantrace', 0, 0,
	{
	"event 201" => -1,
	"event 202" => -1,
	"event 203" => -1,
	"event 204" => -1,
	"event 205" => -1,
	"event 206" => -1,
	"event 207" => -1,
	"event 208" => -1,
	"event 209" => -1,
	"event 210" => -1,
	"event 261" => -1
	},
	{
	"event 262" => -1,
	},
	undef );

Irssi::signal_add('redir scantrace', 'event_trace_line');
Irssi::signal_add('redir scantracestop', 'event_trace_end');


Irssi::signal_add_last('server event',     'event_serverevent');
Irssi::signal_add_last('message public',   'event_public');

# ---------------------------------------------------------------------
# Add settings

Irssi::settings_add_bool('dss', 'dss_enable',               0);
Irssi::settings_add_bool('dss', 'dss_enable_query_public',  0);
Irssi::settings_add_bool('dss', 'dss_enable_remote',        1);
Irssi::settings_add_bool('dss', 'dss_enable_log_hosts',     0);
Irssi::settings_add_bool('dss', 'dss_enable_log_klines',    1);

Irssi::settings_add_str('dss', 'dss_server_type', 'efnet');
Irssi::settings_add_int('dss', 'dss_time', 1440);
Irssi::settings_add_str('dss', 'dss_reason', "dns pollution. read http://spamcalc.net/dnsspam.php and use another host.");

Irssi::settings_add_str('dss', 'dss_fetch_location', "http://spamcalc.net/data/");
Irssi::settings_add_bool('dss', 'dss_fetch_enable', 0);
Irssi::settings_add_int('dss', 'dss_fetch_delta', 3600);

Irssi::settings_add_str('dss', 'dss_logfile_hosts', "$irssidir/dss_hosts.log");
Irssi::settings_add_str('dss', 'dss_logfile_klines', "$irssidir/dss_klines.log");

Irssi::settings_add_int('dss', 'dss_threshold_possible', 50);
Irssi::settings_add_int('dss', 'dss_threshold_certain', 100);

#Irssi::settings_add_bool('dss', 'dss_use_tkserv', 1);
#Irssi::settings_add_bool('dss', 'dss_use_external_command', 0);
#Irssi::settings_add_str('dss', 'dss_external_command', "echo \"%spamvalue - %host\" | mail -s dnsspam root\@localhost");

Irssi::settings_add_int('dss', 'dss_netblock_min', 3);
Irssi::settings_add_bool('dss', 'dss_netblock_report_warning', 1);

Irssi::settings_add_str('dss', 'dss_report_channel', '');

Irssi::settings_add_str('dss', 'dss_acl_hostmasks', '');

Irssi::settings_add_int('dss', 'dss_debug_level', 0);

# ---------------------------------------------------------------------
# Initialization

# Init the calculator
$calc = SpamCalc->new();
dprint(0, "Initialized spamcalc object version " . $calc->VERSION);

dprint(0, "Loading spamcalc datafiles...");
$calc->load_datafiles($datafilesdir);

dprint(0, "Make sure %_dss_server_type%_ and %_dss_report_channel%_ are set properly before using this script. Then use %_/set dss_enable ON%_ to enable automatic dnsspam K/G-lines.");
dprint(0, "Use /dss help for help.");
$starttime = time();

# Add the fetch update timer
$fetch_delta = Irssi::settings_get_int("dss_fetch_delta") * 60 * 1000;
if ($fetch_delta < 60000) {
	$fetch_delta = 600000;
}
$delta_handle_fetch = Irssi::timeout_add($fetch_delta, 'delta_fetch', undef);

# ---------------------------------------------------------------------
# EOF

